import api from '@/api.service'
import moment from 'moment'
export const namespaced = true

export const state = () => {
  return {
    branches: [],
    commits: [],
    states: [],
    notificationTime: '0'
  }
}

const SET_COMMITS = 'setCommits'
const SET_BRANCHES = 'setBranches'
const ADD_COMMIT = 'addCommit'
const ADD_STATE = 'addState'
const ADD_BRANCH = 'addBranch'
const UPDATE_BRANCH = 'updateBranch'

export const actions = {
  async fetchCommits({ commit }) {
    const { data } = await api.get('commits')
    commit(SET_COMMITS, data)
  },

  async fetchBranches({ commit }) {
    const { data } = await api.get('branches')
    commit(SET_BRANCHES, data)
  },

  async fetchState({ commit }, id) {
    const { data } = await api.get(`states/${id}`)
    commit(ADD_STATE, data)
  },

  async saveCommit({ commit, state, rootState }, { title, message, branchId, previousCommitId }) {
    const storedCommit = await api.post(`commits`, {
      title,
      message,
      type: 'commit',
      authorId: rootState.auth.user.id,
      branchId,
      time: 15 + state.commits.length * 5,
      previousCommitId,
      timestamp: moment().format('YYYY-MM-DD H:mm:ss')
    })
    const visualizationState = rootState.data[previousCommitId[0]]
    const storedState = await api.post(`states`, visualizationState)
    commit(ADD_STATE, storedState.data)
    commit(ADD_COMMIT, storedCommit.data)
    await api.patch(`branches/${branchId}`, state.branches.find(b => b.id === branchId))
  },

  async saveBranch({ commit, rootState }, { title, description, commits }) {
    const { data } = await api.post(`branches`, {
      title,
      description,
      authorId: rootState.auth.user.id,
      closed: false,
      timestamp: moment().format('YYYY-MM-DD H:mm:ss'),
      commits
    })
    commit(ADD_BRANCH, data)
  },

  async updateBranch({ commit }, branch) {
    const { updatedBranch } = await api.patch(`branches/${branch.id}`, branch)
    commit(UPDATE_BRANCH, updatedBranch)
  },

  setNotificationTime({ commit }, pl) {
    commit('SET_NOTIFICATION_TIME', {
      //slotId: pl.slotId,
      timestamp: pl.timestamp
    })
  }
}

export const mutations = {
  [SET_COMMITS](state, commits) {
    state.commits = commits
  },

  [SET_BRANCHES](state, branches) {
    state.branches = branches
  },

  [ADD_COMMIT](state, commit) {
    state.commits.push(commit)
    state.branches.find(b => b.id === commit.branchId).commits.push(commit.id)
  },

  [ADD_STATE](state, vizualizationState) {
    state.states.push(vizualizationState)
  },

  [ADD_BRANCH](state, branch) {
    state.branches.push(branch)
  },

  [UPDATE_BRANCH](state, branch) {
    state.branches = [...state.branches.filter(b => b.id !== branch.id), branch]
  },

  /**
   * Sets new notification time
   *
   * @param {object} pl payload
   * @param {number}   pl.slotId dataset slot
   * @param {String}   pl.timestamp
   */
  SET_NOTIFICATION_TIME(state, pl) {
    state.notificationTime = pl.timestamp
    //Vue.set(state[pl.slotId], 'notificationTime', pl.timestamp)
  }
}

export const getters = {
  getBranchById: state => branchId => {
    return state.branches.find(b => b.id === branchId)
  },

  getCommitById: state => commitId => {
    return state.commits.find(c => c.id === commitId)
  },

  getStateById: state => id => {
    return state.states.find(c => c.id === id)
  },

  // getNotificationTime: state => slotId => state[slotId].notificationTime
  // eslint-disable-next-line no-unused-vars
  getNotificationTime: state => state.notificationTime
}
